/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

#include "mod_cache.h"
#include "http_conf_globals.h"
#include "http_log.h"
#include "http_main.h"
#include "apr_md5.h"

int file_cache_get_hdrs(cache_req *cq)
{
    if(cq->state & FC_TEMP_FILE && !cq->hdrs)
        cq->hdrs = apr_make_table(cq->el.cache->pool, 20);
    return file_cache_read_hdrs(cq);
}
int file_cache_get_data(cache_req *cq)
{
    if(!cq->fp)
    {
        const char *file = cq->datafile;
        int mode = APR_WRITE | APR_READ | APR_BINARY;
        if(cq->state & FC_TEMP_FILE)
        {
            file = cq->tempfile;
            mode |= APR_CREATE | APR_EXCL;
        }
        if(!file ||
           !(cq->fp = ap_bopenf(cq->el.cache->pool, file, mode, APR_UREAD | APR_UWRITE | APR_GREAD | APR_WREAD)))
        {
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
                         "cache: error opening file %s",
                         file);
            return 0;
        }
    }
    return 1;
}

/* *** THESE NEXT FUNCTIONS WERE STOLEN FROM MOD_PROXY, PROBABY OUGHT TO PUT
       THESE SOMEWHERE BOTH CAN ACCESS, AND MAINTAIN IN ONE PLACE, NAMES HAVE
       BEEN CHANGED JUST SO I DON'T DEPEND ON MOD_PROXY -
       BUT CAN BE BUILT WITH IT */
/*
 * Converts apr_time_t hex digits to a time integer
 */
apr_time_t ap_cache_hex2msec(const char *x)
{
    int i, ch;
    apr_time_t j;
    for (i = 0, j = 0; i < sizeof(j) * 2; i++) {
        ch = x[i];
        j <<= 4;
        if (ap_isdigit(ch))
            j |= ch - '0';
        else if (ap_isupper(ch))
            j |= ch - ('A' - 10);
        else
            j |= ch - ('a' - 10);
    }
    return j;
}

/*
 * Converts a time integer to apr_time_t hex digits
 */
void ap_cache_msec2hex(apr_time_t j, char *y)
{
    int i, ch;

    for (i = (sizeof(j) * 2)-1; i >= 0; i--) {
        ch = j & 0xF;
        j >>= 4;
        if (ch >= 10)
            y[i] = ch + ('A' - 10);
        else
            y[i] = ch + '0';
    }
    y[sizeof(j) * 2] = '\0';
}
/* NOTE: This routine is taken from http_protocol::getline()
 * because the old code found in the proxy module was too
 * difficult to understand and maintain.
 */
/* Get a line of protocol input, including any continuation lines
 * caused by MIME folding (or broken clients) if fold != 0, and place it
 * in the buffer s, of size n bytes, without the ending newline.
 *
 * Returns -1 on error, or the length of s.
 *
 * Note: Because bgets uses 1 char for newline and 1 char for NUL,
 *       the most we can get is (n - 2) actual characters if it
 *       was ended by a newline, or (n - 1) characters if the line
 *       length exceeded (n - 1).  So, if the result == (n - 1),
 *       then the actual input line exceeded the buffer length,
 *       and it would be a good idea for the caller to puke 400 or 414.
 */
static int cache_getline(char *s, int n, BUFF *in, int fold)
{
    char *pos, next;
    int retval;
    int total = 0;

    pos = s;

    do {
        retval = ap_bgets(pos, n, in);     /* retval == -1 if error, 0 if EOF */

        if (retval <= 0)
            return ((retval < 0) && (total == 0)) ? -1 : total;

        /* retval is the number of characters read, not including NUL      */

        n -= retval;            /* Keep track of how much of s is full     */
        pos += (retval - 1);    /* and where s ends                        */
        total += retval;        /* and how long s has become               */

        if (*pos == '\n') {     /* Did we get a full line of input?        */
            *pos = '\0';
            --total;
            ++n;
        }
        else
            return total;       /* if not, input line exceeded buffer size */

        /* Continue appending if line folding is desired and
         * the last line was not empty and we have room in the buffer and
         * the next line begins with a continuation character.
         */
    } while (fold && (retval != 1) && (n > 1)
                  && (next = ap_blookc(in))
                  && ((next == ' ') || (next == '\t')));

    return total;
}

/* These two functions get and put state information into the data file for an ap_cache_el, this state
   informatino will be read and written transparent to clients of this module */
static int file_cache_read_mydata(BUFF *fp, cache_req *cq)
{
    char urlbuff[1034];
    int len, offset=0;

    if(cq->state & FC_READ_MY_DATA || !cq->hdrsfile)
        return 1;

    /* read the data from the cache file */
    /* format
     * date SP expire SP count CRLF
     * dates are stored as hex seconds since 1970
     */
    len = ap_bgets(urlbuff, sizeof urlbuff, fp);
    if (len == -1)
        return -1;
    if (len == 0 || urlbuff[len - 1] != '\n')
        return 0;
    urlbuff[len - 1] = '\0';

    if (!ap_checkmask(urlbuff, "&&&&&&&&&&&&&&&& &&&&&&&&&&&&&&&& &&&&&&&&&&&&&&&&"))
        return 0;

    cq->date = ap_cache_hex2msec(urlbuff + offset);
    offset += (sizeof(cq->date)*2) + 1;
    cq->expire = ap_cache_hex2msec(urlbuff + offset);
    offset += (sizeof(cq->expire)*2) + 1;
    cq->version = ap_cache_hex2msec(urlbuff + offset);
    
    /* check that we have the same URL */
    len = ap_bgets(urlbuff, sizeof urlbuff, fp);
    if (len == -1)
        return 0;
    if (len == 0 || strncmp(urlbuff, "X-NAME: ", 7) != 0 || urlbuff[len - 1] != '\n')
        return 0;
    urlbuff[len - 1] = '\0';
    if (strcmp(urlbuff + 8, cq->el.name) != 0)
        return 0;
    
    cq->state |= FC_READ_MY_DATA;
    return 1;
}

/*
 * Reads headers from a buffer and returns an array of headers.
 * Returns NULL on file error
 * This routine tries to deal with too long lines and continuation lines.
 * @@@: XXX: FIXME: currently the headers are passed thru un-merged. 
 * Is that okay, or should they be collapsed where possible?
 */
int file_cache_read_hdrs(cache_req *cq)
{
    int len;
    char *value, *end;
    char field[MAX_STRING_LEN];
    char buffer[1034];
    BUFF *f;
    if(cq->state & FC_READ_HEADERS)
        return 1;
    
    if(!cq->hdrs) cq->hdrs = apr_make_table(cq->el.cache->pool, 20);

    if(!cq->fp && cq->state & FC_TEMP_FILE) {
        cq->state |= FC_READ_HEADERS | FC_READ_MY_DATA; /* done, because there is nothing to be done :) */
        return 1; 
    }
    
    if(!cq->hdrsfile || !(f = ap_bopenf(cq->el.cache->pool, cq->hdrsfile, APR_READ | APR_BINARY, 0)))
        return 0;
    if(!file_cache_read_mydata(f, cq))
    {
        ap_bclose(f);
        return 0;
    }
    
    /*
     * Read header lines until we get the empty separator line, a read error,
     * the connection closes (EOF), or we timeout.
     */
    while ((len = cache_getline(buffer, sizeof(buffer), f, 1)) > 0) {
        if (!(value = strchr(buffer, ':'))) {     /* Find the colon separator */
            continue;
        }

        *value = '\0';
        ++value;
        /* XXX: RFC2068 defines only SP and HT as whitespace, this test is
         * wrong... and so are many others probably.
         */
        while (ap_isspace(*value))
            ++value;            /* Skip to start of value   */

        /* should strip trailing whitespace as well */
        for (end = &value[strlen(value)-1]; end > value && ap_isspace(*end); --end)
            *end = '\0';

        apr_table_add(cq->hdrs, buffer, value);
        /* the header was too long; at the least we should skip extra data */
        if (len >= sizeof(buffer) - 1) { 
            while ((len = cache_getline(field, MAX_STRING_LEN, f, 1))
                   >= MAX_STRING_LEN - 1) {
                /* soak up the extra data */
            }
            if (len == 0) /* time to exit the larger loop as well */
                break;
        }
    }
    ap_bclose(f);
    cq->state |= FC_READ_HEADERS;
    return 1;
}

static int file_cache_write_mydata(BUFF *fp, cache_req *cq)
{
    char buff[sizeof(apr_time_t) * 2];
    cache_server_conf *conf = (cache_server_conf *)
        ap_get_module_config(cq->el.cache->server->module_config, &cache_module);
    apr_time_t now, lmod;
    const char *expire, *lmods;
    char *dates;

    if(!cq->hdrs && !file_cache_get_hdrs(cq))
    {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
                     "cache: error getting headers for mydata");
        return 0;
    }

    now = apr_now();
    /* 1) date generation */
    dates = (char *)apr_table_get(cq->hdrs, "Date");
    if (dates != NULL)
        cq->date = ap_parseHTTPdate(dates);
    if (!dates || cq->date == BAD_DATE) {    /* No, or bad date */
        /* no date header! */
        /* add one; N.B. use the time _now_ rather than when we were checking the cache */
        cq->date = now;
        apr_rfc822_date(dates = apr_palloc(cq->el.cache->pool, AP_RFC822_DATE_LEN), now);
        apr_table_set(cq->hdrs, "Date", dates);
        ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, NULL,
                     "Added date header");
    }
    /* 2) last mod generation */
    lmods = apr_table_get(cq->hdrs, "Last-Modified");
    if (lmods != NULL) {
        lmod = ap_parseHTTPdate(lmods);
        if (lmod == BAD_DATE) 
            lmods = NULL; /* kill last modified date */
    }
    if(!lmods || (lmod != BAD_DATE && lmod > cq->date))
    {
        lmod = cq->date;
        ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, NULL,
                     "Last modified is in the future, replacing with now");
    }
    /* 3) expiration generation */
    expire = apr_table_get(cq->hdrs, "Expires");
    if (expire != NULL)
        cq->expire = ap_parseHTTPdate(expire);
    if(!expire || cq->expire == BAD_DATE)
    {
        /* so we now have the expiry date */
        /* if no expiry date then
         *   if lastmod
         *      expiry date = now + min((date - lastmod) * factor, maxexpire)
         *   else
         *      expire date = now + defaultexpire
         */
        if (lmod != BAD_DATE) {
            apr_time_t x = ((apr_time_t) conf->lmfactor) * (cq->date - lmod);
            if (x > conf->maxexpire)
                x = conf->maxexpire;
            cq->expire = now + x;
        }
        else
            cq->expire = now + conf->defaultexpire;
        ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, NULL,
                         "Expiry date calculated %ld", (long)(cq->expire));
    }

    ap_cache_msec2hex(cq->date, buff);
    ap_bvputs(fp, buff, " ", NULL);
    ap_cache_msec2hex(cq->expire, buff);
    ap_bvputs(fp, buff, " ", NULL);
    ap_cache_msec2hex(cq->version++, buff);
    ap_bvputs(fp, buff, "\n", NULL);

    if (ap_bvputs(fp, "X-NAME: ", cq->el.name, "\n", NULL) == -1) {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
                     "cache: error writing (my_data) in cache");
        return 0;
    }
    return 1;
}
int file_cache_write_hdrs(cache_req *cq)
{
    BUFF *fp;
    cache_server_conf *conf = (cache_server_conf *)
        ap_get_module_config(cq->el.cache->server->module_config, &cache_module);

    if(cq->state & FC_READ_HEADERS || cq->state & FC_READ_MY_DATA)
    {
        if(!cq->hdrsfile) cq->hdrsfile = header_file(cq->el.cache, cq->el.name);
        if(unlink(cq->hdrsfile)) /* if we can remove it, we clearly don't have to build the dirs */
            mkdir_structure(cq->hdrsfile, conf->root);
        if(!(fp = ap_bopenf(cq->el.cache->pool, cq->hdrsfile, APR_WRITE | APR_CREATE | APR_BINARY | APR_EXCL,
                            APR_UREAD | APR_UWRITE | APR_GREAD | APR_WREAD))) {
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
                         "cache: error creating cache header file %s",
                         cq->hdrsfile);
            return errno;
        }
        if(cq->state & FC_READ_MY_DATA)
            file_cache_write_mydata(fp, cq);
        if(cq->hdrs)
        {
            int i;
            apr_table_entry_t *elts = (apr_table_entry_t *) ap_table_elts(cq->hdrs)->elts;
            for (i = 0; i < ap_table_elts(cq->hdrs)->nelts; ++i) {
                if (elts[i].key != NULL) {
                    ap_bvputs(fp, elts[i].key, ": ", elts[i].val, CRLF, NULL);
                }
            }
        }
        ap_bclose(fp); /* flush and close */
    }
    return 1;
}

int mkdir_structure(char *file, const char *root)
{
    int rootlen = root ? strlen(root) : 0;
    char *p;
    for (p = file + rootlen + 1;;) {
        p = strchr(p, '/');
        if (!p)
            break;
        *p = '\0';
#ifdef WIN32
        if (mkdir(file) < 0 && errno != EEXIST)
#elif defined(__TANDEM)
            if (mkdir(file, S_IRWXU | S_IRWXG | S_IRWXO) < 0 && errno != EEXIST)
#else
                if (mkdir(file, S_IREAD | S_IWRITE | S_IEXEC) < 0 && errno != EEXIST)
#endif /* WIN32 */
                    ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
                                 "cache: error creating cache directory %s",
                                 file);
        *p = '/';
        ++p;
    }
    return 1;
}

/* ********* FIGURES OUT FILENAMES FOR CACHE_ELEMENT */
#ifdef CASE_BLIND_FILESYSTEM
/*
 * On some platforms, the file system is NOT case sensitive. So, a == A
 * need to map to smaller set of characters
 */
static void ap_cache_hash(const char *it, char *val, int ndepth, int nlength)
{
    ap_md5_ctx_t context;
    unsigned char digest[16];
    char tmp[26];
    int i, k, d;
    unsigned int x;
    static const char enc_table[32] = "abcdefghijklmnopqrstuvwxyz012345";

    apr_MD5Init(&context);
    apr_MD5Update(&context, (const unsigned char *) it, strlen(it));
    apr_MD5Final(digest, &context);

/* encode 128 bits as 26 characters, using a modified uuencoding */
/* the encoding is 5 bytes -> 8 characters
 * i.e. 128 bits is 3 x 5 bytes + 1 byte -> 3 * 8 characters + 2 characters
 */
    for (i = 0, k = 0; i < 15; i += 5) {
    x = (digest[i] << 24) | (digest[i + 1] << 16) | (digest[i + 2] << 8) | digest[i + 3];
    tmp[k++] = enc_table[x >> 27];
    tmp[k++] = enc_table[(x >> 22) & 0x1f];
    tmp[k++] = enc_table[(x >> 17) & 0x1f];
    tmp[k++] = enc_table[(x >> 12) & 0x1f];
    tmp[k++] = enc_table[(x >> 7) & 0x1f];
    tmp[k++] = enc_table[(x >> 2) & 0x1f];
    x = ((x & 0x3) << 8) | digest[i + 4];
    tmp[k++] = enc_table[x >> 5];
    tmp[k++] = enc_table[x & 0x1f];
    }
/* one byte left */
    x = digest[15];
    tmp[k++] = enc_table[x >> 3];    /* use up 5 bits */
    tmp[k++] = enc_table[x & 0x7];
    /* now split into directory levels */

    for (i = k = d = 0; d < ndepth; ++d) {
    memcpy(&val[i], &tmp[k], nlength);
    k += nlength;
    val[i + nlength] = '/';
    i += nlength + 1;
    }
    memcpy(&val[i], &tmp[k], 26 - k);
    val[i + 26 - k] = '\0';
}

#else
static void ap_cache_hash(const char *it, char *val, int ndepth, int nlength)
{
    ap_md5_ctx_t context;
    unsigned char digest[16];
    char tmp[22];
    int i, k, d;
    unsigned int x;
#if defined(AIX) && defined(__ps2__)
    /* Believe it or not, AIX 1.x does not allow you to name a file '@',
     * so hack around it in the encoding. */
    static const char enc_table[64] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_%";
#else
    static const char enc_table[64] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_@";
#endif

    apr_MD5Init(&context);
    apr_MD5Update(&context, (const unsigned char *) it, strlen(it));
    apr_MD5Final(digest, &context);

/* encode 128 bits as 22 characters, using a modified uuencoding */
/* the encoding is 3 bytes -> 4 characters
 * i.e. 128 bits is 5 x 3 bytes + 1 byte -> 5 * 4 characters + 2 characters
 */
    for (i = 0, k = 0; i < 15; i += 3) {
    x = (digest[i] << 16) | (digest[i + 1] << 8) | digest[i + 2];
    tmp[k++] = enc_table[x >> 18];
    tmp[k++] = enc_table[(x >> 12) & 0x3f];
    tmp[k++] = enc_table[(x >> 6) & 0x3f];
    tmp[k++] = enc_table[x & 0x3f];
    }
/* one byte left */
    x = digest[15];
    tmp[k++] = enc_table[x >> 2];    /* use up 6 bits */
    tmp[k++] = enc_table[(x << 4) & 0x3f];
    /* now split into directory levels */

    for (i = k = d = 0; d < ndepth; ++d) {
    memcpy(&val[i], &tmp[k], nlength);
    k += nlength;
    val[i + nlength] = '/';
    i += nlength + 1;
    }
    memcpy(&val[i], &tmp[k], 22 - k);
    val[i + 22 - k] = '\0';
}

#endif /* CASE_BLIND_FILESYSTEM */
static char *generate_name(ap_cache_handle_t *h, const char *name)
{
    char hashfile[66], *filebase;
    cache_server_conf *conf = (cache_server_conf *)    ap_get_module_config(h->server->module_config, &cache_module);
    ap_cache_hash(name, hashfile, conf->dirlevels, conf->dirlength);
    filebase = apr_pstrcat(h->pool, conf->root, "/", hashfile, "%s", NULL);
    return filebase;
}
char *header_file(ap_cache_handle_t *h, const char *name)
{
    return apr_psprintf(h->pool, generate_name(h, name), CACHE_HEADER_SUFFIX);
}
char *data_file(ap_cache_handle_t *h, const char *name)
{
    return apr_psprintf(h->pool, generate_name(h, name), CACHE_DATA_SUFFIX);
}
